@@ -95,7 +95,7 @@ use vars qw{
   @site_rules_path
 };
 
-$VERSION = "3.002004";      # update after release (same format as perl $])
+$VERSION = "3.002005";      # update after release (same format as perl $])
 # $IS_DEVEL_BUILD = 1;        # change for release versions
 
 # Used during the prerelease/release-candidate part of the official release
@@ -106,11 +106,11 @@ $VERSION = "3.002004";      # update after release (same format as perl $])
 @ISA = qw();
 
 # SUB_VERSION is now just <yyyy>-<mm>-<dd>
-$SUB_VERSION = (split(/\s+/,'$LastChangedDate: 2008-01-01 20:59:55 +0000 (Tue, 01 Jan 2008) $ updated by SVN'))[1];
+$SUB_VERSION = (split(/\s+/,'$LastChangedDate: 2008-06-10 09:13:55 +0000 (Tue, 10 Jun 2008) $ updated by SVN'))[1];
 
 if (defined $IS_DEVEL_BUILD && $IS_DEVEL_BUILD) {
   push(@EXTRA_VERSION,
-       ('r' . qw{$LastChangedRevision: 607910 $ updated by SVN}[1]));
+       ('r' . qw{$LastChangedRevision: 666026 $ updated by SVN}[1]));
 }
 
 sub Version {
@@ -1706,6 +1706,7 @@ sub sed_path {
   $path =~ s/__def_rules_dir__/$self->{DEF_RULES_DIR} || ''/ges;
   $path =~ s{__prefix__}{$self->{PREFIX} || $Config{prefix} || '/usr'}ges;
   $path =~ s{__userstate__}{$self->get_and_create_userstate_dir() || ''}ges;
+  $path =~ s{__perl_major_ver__}{$self->get_perl_major_version()}ges;
   $path =~ s/__version__/${VERSION}/gs;
   $path =~ s/^\~([^\/]*)/$self->expand_name($1)/es;
 
@@ -1714,6 +1715,12 @@ sub sed_path {
   return $path;
 }
 
+sub get_perl_major_version {
+  my $self = shift;
+  $] =~ /^(\d\.\d\d\d)/ or die "bad perl ver $]";
+  return $1;
+}
+
 sub first_existing_path {
   my $self = shift;
   my $path;
