@@ -279,7 +279,11 @@ sub tie_db_writable {
     my $name = $path.'_'.$dbname;
     my $db_var = 'db_'.$dbname;
     dbg("bayes: tie-ing to DB file R/W $name");
-    tie %{$self->{$db_var}},$self->DBM_MODULE,$name, O_RDWR|O_CREAT,
+
+    ($self->DBM_MODULE eq 'DB_File') and
+         Mail::SpamAssassin::Util::avoid_db_file_locking_bug ($name);
+
+    tie %{$self->{$db_var}},$self->DBM_MODULE, $name, O_RDWR|O_CREAT,
 		 (oct($main->{conf}->{bayes_file_mode}) & 0666)
        or goto failed_to_tie;
   }
@@ -287,7 +291,7 @@ sub tie_db_writable {
 
   # set our cache to what version DB we're using
   $self->{db_version} = ($self->get_storage_variables())[6];
-  # don't bother printing this if the db was not found since it is bogus anyway
+  # don't bother printing this unless found since it would be bogus anyway
   dbg("bayes: found bayes db version ".$self->{db_version}) if ($found);
 
   # figure out if we can read the current DB and if we need to do a
@@ -590,7 +594,7 @@ sub untie_db {
     my $db_var = 'db_'.$dbname;
 
     if (exists $self->{$db_var}) {
-      dbg("bayes: untie-ing $db_var");
+      # dbg("bayes: untie-ing $db_var");
       untie %{$self->{$db_var}};
       delete $self->{$db_var};
     }
@@ -1123,8 +1127,12 @@ sub cleanup {
 
     # argh, write failure, give up
     if (!defined $len || $len < 0) {
-      $len = 0 unless (defined $len);
-      warn "bayes: write failed to Bayes journal $path ($len of $nbytes)!\n";
+      my $err = '';
+      if (!defined $len) {
+	$len = 0;
+	$err = "  ($!)";
+      }
+      warn "bayes: write failed to Bayes journal $path ($len of $nbytes)!$err\n";
       last;
     }
 
@@ -1511,7 +1519,7 @@ sub clear_database {
     my $db_var = 'db_'.$dbname;
 
     if (exists $self->{$db_var}) {
-      dbg("bayes: untie-ing $db_var");
+      # dbg("bayes: untie-ing $db_var");
       untie %{$self->{$db_var}};
       delete $self->{$db_var};
     }
