@@ -107,7 +107,7 @@ sub load {
    };
 
    if ($@) {
-     warn "ldap: failed to load user scores from LDAP server, ignored\n";
+     warn "ldap: failed to load user scores from LDAP server, ignored ($@)\n";
    }
 }
 
@@ -130,6 +130,7 @@ sub load_with_ldap {
   my @attr   = $uri->attributes;
   my $scope  = $uri->scope;
   my $filter = $uri->filter;
+  my $scheme = $uri->scheme;
   my %extn   = $uri->extensions; # unused
 
   $filter =~ s/__USERNAME__/$username/g;
@@ -154,7 +155,10 @@ sub load_with_ldap {
 
   my $f_attribute = $attr[0];
 
-  my $ldap = Net::LDAP->new ("$host:$port", onerror => "warn");
+  my $ldap = Net::LDAP->new ("$host:$port",
+                onerror => "warn",
+                scheme => $scheme);
+
   if (!defined($ldapuser) && !defined($ldappass)) {
     $ldap->bind;
   } else {
