@@ -107,6 +107,7 @@ sub do_rbl_lookup {
 
     my $ent = {
       key => $key,
+      zone => $host,  # may serve to fetch other per-zone settings
       type => "DNSBL-".$type,
       sets => [ ],  # filled in below
       rules => [ ], # filled in below
@@ -114,10 +115,9 @@ sub do_rbl_lookup {
     };
 
     my $id = $self->{resolver}->bgsend($host, $type, undef, sub {
-        my $pkt = shift;
-        my $id = shift;
+        my ($pkt, $id, $timestamp) = @_;
         $self->process_dnsbl_result($ent, $pkt);
-        $self->{async}->report_id_complete($id);
+        $self->{async}->report_id_complete($id,$key,$timestamp);
       });
 
     $ent->{id} = $id;     # tie up the loose end
@@ -156,16 +156,16 @@ sub do_dns_lookup {
 
   my $ent = {
     key => $key,
+    zone => $host,  # may serve to fetch other per-zone settings
     type => "DNSBL-".$type,
     rules => [ $rule ],
     # id is filled in after we send the query below
   };
 
   my $id = $self->{resolver}->bgsend($host, $type, undef, sub {
-      my $pkt = shift;
-      my $id = shift;
+      my ($pkt, $id, $timestamp) = @_;
       $self->process_dnsbl_result($ent, $pkt);
-      $self->{async}->report_id_complete($id);
+      $self->{async}->report_id_complete($id,$key,$timestamp);
     });
 
   $ent->{id} = $id;     # tie up the loose end
@@ -328,75 +328,78 @@ sub process_dnsbl_set {
 sub harvest_until_rule_completes {
   my ($self, $rule) = @_;
 
-  return if !defined $self->{async}->get_last_start_lookup_time();
+  dbg("dns: harvest_until_rule_completes");
+  my $result = 0;
+  my $total_waiting_time = 0;
 
-  my $deadline = $self->{conf}->{rbl_timeout} + $self->{async}->get_last_start_lookup_time();
-  my $now = time;
+  for (my $first=1;  ; $first=0) {
+    # complete_lookups() may call completed_callback(), which may
+    # call start_lookup() again (like in Plugin::URIDNSBL)
+    my ($alldone,$anydone,$waiting_time) =
+      $self->{async}->complete_lookups($first ? 0 : 1.0,  1);
+    $total_waiting_time += $waiting_time;
 
-  my @left = $self->{async}->get_pending_lookups();
-  my $total = scalar @left;
-
-  while (($now < $deadline) && !$self->{async}->complete_lookups(1))
-  {
-    if ($self->is_rule_complete($rule)) {
-      return 1;
-    }
+    $result = 1  if $self->is_rule_complete($rule);
+    last  if $result || $alldone;
 
+    dbg("dns: harvest_until_rule_completes - check_tick");
     $self->{main}->call_plugins ("check_tick", { permsgstatus => $self });
-    @left = $self->{async}->get_pending_lookups();
-
-    # dynamic timeout
-    my $dynamic = (int($self->{conf}->{rbl_timeout}
-                      * (1 - (($total - scalar @left) / $total) ** 2) + 0.5)
-                  + $self->{async}->get_last_start_lookup_time());
-    $deadline = $dynamic if ($dynamic < $deadline);
-    $now = time;
   }
+  dbg("dns: timing: %.3f s sleeping in harvest_until_rule_completes",
+      $total_waiting_time)  if $total_waiting_time > 0;
+
+  return $result;
 }
 
 sub harvest_dnsbl_queries {
   my ($self) = @_;
 
-  return if !defined $self->{async}->get_last_start_lookup_time();
+  dbg("dns: harvest_dnsbl_queries");
+  my $total_waiting_time = 0;
 
-  my $deadline = $self->{conf}->{rbl_timeout} + $self->{async}->get_last_start_lookup_time();
-  my $now = time;
+  for (my $first=1;  ; $first=0) {
 
-  my @left = $self->{async}->get_pending_lookups();
-  my $total = scalar @left;
+    # complete_lookups() may call completed_callback(), which may
+    # call start_lookup() again (like in Plugin::URIDNSBL)
 
-  while (($now < $deadline) && !$self->{async}->complete_lookups(1))
-  {
+    # the first time around we specify a 0 timeout, which gives
+    # complete_lookups a chance to ripe any available results and
+    # abort overdue requests, without needlessly waiting for more
+
+    my ($alldone,$anydone,$waiting_time) =
+      $self->{async}->complete_lookups($first ? 0 : 1.0,  1);
+    $total_waiting_time += $waiting_time;
+
+    last  if $alldone;
+
+    dbg("dns: harvest_dnsbl_queries - check_tick");
     $self->{main}->call_plugins ("check_tick", { permsgstatus => $self });
-    @left = $self->{async}->get_pending_lookups();
-
-    # dynamic timeout
-    my $dynamic = (int($self->{conf}->{rbl_timeout}
-                      * (1 - (($total - scalar @left) / $total) ** 2) + 0.5)
-                  + $self->{async}->get_last_start_lookup_time());
-    $deadline = $dynamic if ($dynamic < $deadline);
-    $now = time;    # and loop again
   }
 
-  dbg("dns: success for " . ($total - @left) . " of $total queries");
-
-  # timeouts
-  @left = $self->{async}->get_pending_lookups();
-  for my $query (@left) {
-    my $string = '';
-    if (defined @{$query->{sets}}) {
-      $string = join(",", grep defined, @{$query->{sets}});
-    }
-    elsif (defined @{$query->{rules}}) {
-      $string = join(",", grep defined, @{$query->{rules}});
-    }
-    my $delay = time - $self->{async}->get_last_start_lookup_time();
-    dbg("dns: timeout for $string after $delay seconds");
-  }
-
-  # and explicitly abort anything left
+  # explicitly abort anything left
   $self->{async}->abort_remaining_lookups();
+  $self->{async}->log_lookups_timing();
   $self->mark_all_async_rules_complete();
+  dbg("dns: timing: %.3f s sleeping in harvest_dnsbl_queries",
+      $total_waiting_time)  if $total_waiting_time > 0;
+  1;
+}
+
+# collect and process whatever DNS responses have already arrived,
+# don't waste time waiting for more, don't poll too often.
+# don't abort any queries even if overdue, 
+sub harvest_completed_queries {
+  my ($self) = @_;
+
+  # don't bother collecting responses too often
+  my $last_poll_time = $self->{async}->last_poll_responses_time();
+  return if defined $last_poll_time && time - $last_poll_time < 0.1;
+
+  my ($alldone,$anydone) = $self->{async}->complete_lookups(0, 0);
+  if ($anydone) {
+    dbg("dns: harvested completed queries");
+#   $self->{main}->call_plugins ("check_tick", { permsgstatus => $self });
+  }
 }
 
 sub set_rbl_tag_data {
@@ -432,6 +435,12 @@ sub load_resolver {
   return $self->{resolver}->load_resolver();
 }
 
+sub clear_resolver {
+  my ($self) = @_;
+  $self->{main}->{resolver}->{res} = undef;
+  return 0;
+}
+
 sub lookup_ns {
   my ($self, $dom) = @_;
 
@@ -454,11 +463,12 @@ sub lookup_ns {
 	}
       }
       $nsrecords = $self->{dnscache}->{NS}->{$dom} = [ @nses ];
-    };
-    if ($@) {
-      dbg("dns: NS lookup failed horribly, perhaps bad resolv.conf setting? ($@)");
+      1;
+    } or do {
+      my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
+      dbg("dns: NS lookup failed horribly, perhaps bad resolv.conf setting? ($eval_stat)");
       return undef;
-    }
+    };
   }
 
   $nsrecords;
@@ -488,11 +498,12 @@ sub lookup_mx {
       }
 
       $mxrecords = $self->{dnscache}->{MX}->{$dom} = [ @ips ];
-    };
-    if ($@) {
-      dbg("dns: MX lookup failed horribly, perhaps bad resolv.conf setting? ($@)");
+      1;
+    } or do {
+      my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
+      dbg("dns: MX lookup failed horribly, perhaps bad resolv.conf setting? ($eval_stat)");
       return undef;
-    }
+    };
   }
 
   $mxrecords;
@@ -546,12 +557,12 @@ sub lookup_ptr {
       }
 
       $name = $self->{dnscache}->{PTR}->{$dom} = $name;
-    };
-
-    if ($@) {
-      dbg("dns: PTR lookup failed horribly, perhaps bad resolv.conf setting? ($@)");
+      1;
+    } or do {
+      my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
+      dbg("dns: PTR lookup failed horribly, perhaps bad resolv.conf setting? ($eval_stat)");
       return undef;
-    }
+    };
   }
   dbg("dns: PTR for '$dom': '$name'");
 
@@ -588,12 +599,12 @@ sub lookup_a {
 	}
       }
       $self->{dnscache}->{A}->{$name} = [ @addrs ];
-    };
-
-    if ($@) {
-      dbg("dns: A lookup failed horribly, perhaps bad resolv.conf setting? ($@)");
+      1;
+    } or do {
+      my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
+      dbg("dns: A lookup failed horribly, perhaps bad resolv.conf setting? ($eval_stat)");
       return undef;
-    }
+    };
   }
 
   dbg("dns: A records for '$name': ".join (' ', @addrs));
@@ -613,7 +624,8 @@ sub is_dns_available {
   # working DNS and our check interval time has passed
   if ($dnsopt eq "test" && $diff > $dnsint) {
     $IS_DNS_AVAILABLE = undef;
-    dbg("dns: is_dns_available() last checked $diff seconds ago; re-checking");
+    dbg("dns: is_dns_available() last checked %.1f seconds ago; re-checking",
+        $diff);
   }
 
   return $IS_DNS_AVAILABLE if (defined $IS_DNS_AVAILABLE);
@@ -629,12 +641,6 @@ sub is_dns_available {
   # DNS if we're only supposed to be looking at local tests.
   goto done if ($self->{main}->{local_tests_only});
 
-  if ($dnsopt eq "yes") {
-    $IS_DNS_AVAILABLE = 1;
-    dbg("dns: dns_available set to yes in config file, skipping test");
-    return $IS_DNS_AVAILABLE;
-  }
-
   # Check version numbers - runtime check only
   if (defined $Net::DNS::VERSION) {
     if (Mail::SpamAssassin::Util::am_running_on_windows()) {
@@ -651,8 +657,25 @@ sub is_dns_available {
     }
   }
 
+  $self->clear_resolver();
   goto done unless $self->load_resolver();
 
+  my @nameservers = $self->{resolver}->nameservers();
+
+  # optionally shuffle the list of nameservers to distribute the load
+  if ($self->{conf}->{dns_options}->{rotate}) {
+    Mail::SpamAssassin::Util::fisher_yates_shuffle(\@nameservers);
+    dbg("dns: shuffled NS list: ".join(", ", @nameservers));
+    $self->{resolver}->nameservers(@nameservers);
+    $self->{resolver}->connect_sock();
+  }
+
+  if ($dnsopt eq "yes") {
+    $IS_DNS_AVAILABLE = 1;
+    dbg("dns: dns_available set to yes in config file, skipping test");
+    return $IS_DNS_AVAILABLE;
+  }
+
   if ($dnsopt =~ /test:\s+(.+)$/) {
     my $servers=$1;
     dbg("dns: servers: $servers");
@@ -662,15 +685,11 @@ sub is_dns_available {
     @domains = @EXISTING_DOMAINS;
   }
 
-  # TODO: retry every now and again if we get this far, but the
-  # next test fails?  could be because the ethernet cable has
-  # simply fallen out ;)
-
   # Net::DNS::Resolver scans a list of nameservers when it does a foreground query
   # but only uses the first in a background query like we use.
   # Try the different nameservers here in case the first one is not woorking
-  
-  my @nameservers = $self->{resolver}->nameservers();
+
+  my @good_nameservers = ();
   dbg("dns: testing resolver nameservers: ".join(", ", @nameservers));
   my $ns;
   while( $ns  = shift(@nameservers)) {
@@ -682,6 +701,7 @@ sub is_dns_available {
         if (scalar @$result > 0) {
           dbg("dns: NS lookup of $domain using $ns succeeded => DNS available (set dns_available to override)");
           $IS_DNS_AVAILABLE = 1;
+          push(@good_nameservers, $ns);
           last;
         }
         else {
@@ -695,13 +715,18 @@ sub is_dns_available {
         last; 
       }
     }
-    last if $IS_DNS_AVAILABLE;
-    dbg("dns: NS lookups failed, removing nameserver $ns from list");
     $self->{resolver}->nameservers(@nameservers);
     $self->{resolver}->connect_sock(); # reconnect socket to new nameserver
   }
 
-  dbg("dns: all NS queries failed => DNS unavailable (set dns_available to override)") if ($IS_DNS_AVAILABLE == 0);
+  if ($IS_DNS_AVAILABLE == 1)
+  {
+    dbg("dns: NS list: ".join(", ", @good_nameservers));
+    $self->{resolver}->nameservers(@good_nameservers);
+    $self->{resolver}->connect_sock();
+  } else {
+    dbg("dns: all NS queries failed => DNS unavailable (set dns_available to override)");
+  }
 
 done:
   # jm: leaving this in!
