@@ -42,13 +42,20 @@ use bytes;
 use Mail::SpamAssassin;
 use Mail::SpamAssassin::Logger;
 
+use Socket;
 use IO::Socket::INET;
-use Errno qw(EINVAL EADDRINUSE);
+use Errno qw(EADDRINUSE EACCES);
 
-use constant HAS_SOCKET_INET6 => eval { require IO::Socket::INET6 };
+use constant HAS_SOCKET_INET6 => eval { require IO::Socket::INET6; };
 
 our @ISA = qw();
 
+# Load Time::HiRes if it's available
+BEGIN {
+  eval { require Time::HiRes };
+  Time::HiRes->import( qw(time) ) unless $@;
+}
+
 ###########################################################################
 
 sub new {
@@ -81,13 +88,41 @@ sub load_resolver {
 
   if (defined $self->{res}) { return 1; }
   $self->{no_resolver} = 1;
-
+  # force only ipv4 if no IO::Socket::INET6 or ipv6 doesn't work
+  # to be safe test both ipv6 and ipv4 addresses in INET6
+  my $force_ipv4 = (!HAS_SOCKET_INET6) || $self->{main}->{force_ipv4} ||
+    !eval {
+      my $sock6 = IO::Socket::INET6->new(
+                                         LocalAddr => "::",
+                                         Proto     => 'udp',
+                                         );
+      if ($sock6) {
+        $sock6->close();
+        1;
+      }
+    } ||
+    !eval {
+      my $sock6 = IO::Socket::INET6->new(
+                                         LocalAddr => "0.0.0.0",
+                                         PeerAddr => "0.0.0.0",
+					 PeerPort => 53,
+                                         Proto     => 'udp',
+                                         );
+      if ($sock6) {
+        $sock6->close();
+        1;
+      }
+    };
+  
   eval {
     require Net::DNS;
-    $self->{res} = Net::DNS::Resolver->new;
+    # force_v4 is set in new() to avoid error in older versions of Net::DNS that don't have it
+    # other options are set by function calls so a typo or API change will cause an error here
+    $self->{res} = Net::DNS::Resolver->new(force_v4 => $force_ipv4);
     if (defined $self->{res}) {
       $self->{no_resolver} = 0;
-      $self->{retry} = 1;               # retries for non-nackgrounded query
+      $self->{force_ipv4} = $force_ipv4;
+      $self->{retry} = 1;               # retries for non-backgrounded query
       $self->{retrans} = 3;   # initial timeout for "non-backgrounded" query run in background
       $self->{res}->retry(1);           # If it fails, it fails
       $self->{res}->retrans(0);         # If it fails, it fails
@@ -101,6 +136,7 @@ sub load_resolver {
     1;
   };   #  or warn "dns: eval failed: $@ $!\n";
 
+  dbg("dns: no ipv6") if $force_ipv4;
   dbg("dns: is Net::DNS::Resolver available? " .
        ($self->{no_resolver} ? "no" : "yes"));
   if (!$self->{no_resolver} && defined $Net::DNS::VERSION) {
@@ -123,7 +159,7 @@ sub get_resolver {
 
 =item $res->nameservers()
 
-Wrapper for Net::DNS::Reslolver->nameservers to get or set list of nameservers
+Wrapper for Net::DNS::Resolver->nameservers to get or set list of nameservers
 
 =cut
 
@@ -157,20 +193,17 @@ sub connect_sock {
   my $ip64 = IP_ADDRESS;
   my $ip4 = IPV4_ADDRESS;
   my $ns = $self->{res}->{nameservers}[0];
-  my $ipv6 = 0;
+  my $ipv6opt = !($self->{force_ipv4});
 
-  # now, attempt to set the family to AF_INET6 if we can.  Some
-  # platforms don't have it (bug 4412 comment 29)...
-  # also, only set $ipv6 to true if that succeeds.
-  my $family;
-  if (HAS_SOCKET_INET6 && $ns=~/^${ip64}$/o && $ns!~/^${ip4}$/o) {
-    eval '$family = AF_INET6; $ipv6 = 1;';
-  }
-  if (!defined $family) {
-    $family = AF_INET;       # that didn't work ;)
+  # ensure families of src and dest addresses match (bug 4412 comment 29)
+  my $srcaddr;
+  if ($ipv6opt && $ns=~/^${ip64}$/o && $ns!~/^${ip4}$/o) {
+    $srcaddr = "::";
+  } else {
+    $srcaddr = "0.0.0.0";
   }
 
-  dbg("dns: name server: $ns, family: $family, ipv6: $ipv6");
+  dbg("dns: name server: $ns, LocalAddr: $srcaddr");
 
   # find next available unprivileged port (1024 - 65535)
   # starting at a random value to spread out use of ports
@@ -184,10 +217,10 @@ sub connect_sock {
         Proto => 'udp',
         LocalPort => $lport,
         Type => SOCK_DGRAM,
-        Domain => $family,
+        LocalAddr => $srcaddr,
     );
 
-    if (HAS_SOCKET_INET6) {
+    if ($ipv6opt) {
       $sock = IO::Socket::INET6->new(%args);
     } else {
       $sock = IO::Socket::INET->new(%args);
@@ -195,11 +228,9 @@ sub connect_sock {
     $errno = $!;
     if (defined $sock) {  # ok, got it
       last;
-    } elsif ($! == EADDRINUSE) {  # in use, let's try another source port
+    } elsif ($! == EADDRINUSE || $! == EACCES) {  # in use, let's try another source port
       dbg("dns: UDP port $lport already in use, trying another port");
     } else {
-      # did we fail due to the attempted use of an IPv6 nameserver?
-      $self->_ipv6_ns_warning()  if (!$ipv6 && $errno==EINVAL);
       warn "Error creating a DNS resolver socket: $errno";
       goto no_sock;
     }
@@ -209,6 +240,26 @@ sub connect_sock {
     goto no_sock;
   }
 
+  eval {
+    my($bufsiz,$newbufsiz);
+    $bufsiz = $sock->sockopt(Socket::SO_RCVBUF)
+      or die "Can't get a resolver socket rx buffer size: $!";
+    if ($bufsiz >= 32*1024) {
+      dbg("dns: resolver socket rx buffer size is %d bytes", $bufsiz);
+    } else {
+      $sock->sockopt(Socket::SO_RCVBUF, 32*1024)
+        or die "Can't set a resolver socket rx buffer size: $!";
+      $newbufsiz = $sock->sockopt(Socket::SO_RCVBUF)
+        or die "Can't get a resolver socket rx buffer size: $!";
+      dbg("dns: resolver socket rx buffer size changed from %d to %d bytes",
+          $bufsiz, $newbufsiz);
+    }
+    1;
+  } or do {
+    my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
+    info("dns: socket buffer size error: $eval_stat");
+  };
+
   $self->{sock} = $sock;
   $self->{sock_as_vec} = $self->fhs_to_vec($self->{sock});
   return;
@@ -269,13 +320,14 @@ sub new_dns_packet {
 
     # a bit noisy, so commented by default...
     #dbg("dns: new DNS packet time=".time()." host=$host type=$type id=".$packet->id);
-  };
-
-  if ($@) {
+    1;
+  } or do {
+    my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
     # this can happen if Net::DNS isn't available -- but in this
     # case this function should never be called!
-    warn "dns: cannot create Net::DNS::Packet, but new_dns_packet() was called: $@ $!";
-  }
+    warn "dns: cannot create Net::DNS::Packet, but new_dns_packet() was called: $eval_stat";
+  };
+
   return $packet;
 }
 
@@ -292,7 +344,7 @@ sub _packet_id {
   my $ques = $questions[0];
 
   if (defined $ques) {
-    return $id . $ques->qname . $ques->qtype . $ques->qclass;
+    return join '/', $id, $ques->qname, $ques->qtype, $ques->qclass;
   } else {
     # odd.  this should not happen, but clearly some DNS servers
     # can return something that Net::DNS interprets as having no
@@ -314,18 +366,19 @@ sub reference C<$cb> will be called.
 Note that C<$type> and C<$class> may be C<undef>, in which case they
 will default to C<A> and C<IN>, respectively.
 
-The callback sub will be called with two arguments -- the packet that was
-delivered and an id string that fingerprints the query packet and the expected reply.
-It is expected that a closure callback be used, like so:
+The callback sub will be called with three arguments -- the packet that was
+delivered, and an id string that fingerprints the query packet and the expected
+reply. The third argument is a timestamp (Unix time, floating point), captured
+at the time the packet was collected. It is expected that a closure callback
+be used, like so:
 
   my $id = $self->{resolver}->bgsend($host, $type, undef, sub {
-        my $reply = shift;
-        my $reply_id = shift;
+        my ($reply, $reply_id, $timestamp) = @_;
         $self->got_a_reply ($reply, $reply_id);
       });
 
-The callback can ignore the reply as an invalid packet sent to the listening port
-if the reply id does not match the return value from bgsend.
+The callback can ignore the reply as an invalid packet sent to the listening
+port if the reply id does not match the return value from bgsend.
 
 =cut
 
@@ -333,6 +386,8 @@ sub bgsend {
   my ($self, $host, $type, $class, $cb) = @_;
   return if $self->{no_resolver};
 
+  $self->{send_timed_out} = 0;
+
   my $pkt = $self->new_dns_packet($host, $type, $class);
 
   $self->connect_sock_if_reqd();
@@ -358,53 +413,62 @@ sub poll_responses {
   my ($self, $timeout) = @_;
   return if $self->{no_resolver};
   return if !$self->{sock};
+  my $cnt = 0;
+  my $waiting_time = 0;
 
   my $rin = $self->{sock_as_vec};
   my $rout;
-  my ($nfound, $timeleft) = select($rout=$rin, undef, undef, $timeout);
 
-  if (!defined $nfound || $nfound < 0) {
-    warn "dns: select failed: $!";
-    return;
-  }
-
-  if ($nfound == 0) {
-    return 0;           # nothing's ready yet
-  }
-
-  my $packet = $self->{res}->bgread($self->{sock});
-  my $err = $self->{res}->errorstring;
-
-  if (defined $packet &&
-      defined $packet->header &&
-      defined $packet->question &&
-      defined $packet->answer)
-  {
-    my $id = $self->_packet_id($packet);
-
-    my $cb = delete $self->{id_to_callback}->{$id};
-    if (!$cb) {
-      dbg("dns: no callback for id: $id, ignored; packet: ".
-                    ($packet ? $packet->string : "undef"));
-      return 0;
+  for (;;) {
+    my $now_before = time;
+    my ($nfound, $timeleft) = select($rout=$rin, undef, undef, $timeout);
+    if (!defined $nfound || $nfound < 0) {
+      warn "dns: select failed: $!";
+      return;
     }
 
-    $cb->($packet, $id);
-    return 1;
-  }
-  else {
-    dbg("dns: no packet! err=$err packet=".
-                    ($packet ? $packet->string : "undef"));
+    my $now = time;
+    if ($now > $now_before && (!defined($timeout) || $timeout > 0)) {
+      $waiting_time += $now - $now_before;
+    }
+    $timeout = 0;  # next time around collect whatever is available, then exit
+    last  if $nfound == 0;
+
+    my $packet = $self->{res}->bgread($self->{sock});
+    my $err = $self->{res}->errorstring;
+
+    if (defined $packet &&
+        defined $packet->header &&
+        defined $packet->question &&
+        defined $packet->answer)
+    {
+      my $id = $self->_packet_id($packet);
+
+      my $cb = delete $self->{id_to_callback}->{$id};
+      if (!$cb) {
+        dbg("dns: no callback for id: %s, ignored; packet: %s",
+            $id,  $packet ? $packet->string : "undef" );
+      } else {
+        $cb->($packet, $id, $now);
+        $cnt++;
+      }
+    }
+    else {
+      dbg("dns: no packet! err=%s packet=%s",
+          $err,  $packet ? $packet->string : "undef" );
+    }
   }
 
-  return 0;
+  return wantarray ? ($cnt, $waiting_time) : $cnt;
 }
 
 ###########################################################################
 
 =item $res->bgabort()
 
-Call this to release pending requests from memory when aborting backgrounded requests
+Call this to release pending requests from memory, when aborting backgrounded
+requests, or when the scan is complete.
+C<Mail::SpamAssassin::PerMsgStatus::check> calls this before returning.
 
 =cut
 
@@ -429,6 +493,7 @@ sub send {
   my $retries = $self->{retry};
   my $timeout = $retrans;
   my $answerpkt;
+  my $answerpkt_avail = 0;
   for (my $i = 0;
        (($i < $retries) && !defined($answerpkt));
        ++$i, $retrans *= 2, $timeout = $retrans) {
@@ -436,13 +501,15 @@ sub send {
     $timeout = 1 if ($timeout < 1);
     # note nifty use of a closure here.  I love closures ;)
     $self->bgsend($name, $type, $class, sub {
-      $answerpkt = shift;
+      my ($reply, $reply_id, $timestamp) = @_;
+      $answerpkt = $reply; $answerpkt_avail = 1;
     });
 
     my $now = time;
     my $deadline = $now + $timeout;
 
-    while (($now < $deadline) && (!defined($answerpkt))) {
+    while (!$answerpkt_avail) {
+      if ($now >= $deadline) { $self->{send_timed_out} = 1; last }
       $self->poll_responses(1);
       $now = time;
     }
@@ -452,6 +519,26 @@ sub send {
 
 ###########################################################################
 
+=item $res->errorstring()
+
+Little more than a stub for callers expecting this from C<Net::DNS::Resolver>.
+
+If called immediately after a call to $res->send this will return
+C<query timed out> if the $res->send DNS query timed out.  Otherwise 
+C<unknown error or no error> will be returned.
+
+No other errors are reported.
+
+=cut
+
+sub errorstring {
+  my ($self) = @_;
+  return 'query timed out' if $self->{send_timed_out};
+  return 'unknown error or no error';
+}
+
+###########################################################################
+
 =item $res->finish_socket()
 
 Reset socket when done with it.
@@ -477,10 +564,7 @@ Clean up for destruction.
 sub finish {
   my ($self) = @_;
   $self->finish_socket();
-  if (!$self->{no_resolver}) {
-    delete $self->{res};
-  }
-  delete $self->{main};
+  %{$self} = ();
 }
 
 ###########################################################################
@@ -492,8 +576,11 @@ sub fhs_to_vec {
   my $rin = '';
   foreach my $sock (@fhlist) {
     my $fno = fileno($sock);
-    warn "dns: oops! fileno now undef for $sock" unless defined($fno);
-    vec ($rin, $fno, 1) = 1;
+    if (!defined $fno) {
+      warn "dns: oops! fileno now undef for $sock";
+    } else {
+      vec ($rin, $fno, 1) = 1;
+    }
   }
   return $rin;
 }
@@ -506,26 +593,6 @@ sub reinit_post_fork {
   $self->connect_sock();
 }
 
-sub _ipv6_ns_warning {
-  my ($self) = @_;
-
-  # warn about the attempted use of an IPv6 nameserver without
-  # IO::Socket::INET6 installed (bug 4412)
-  my $firstns = $self->{res}->{nameservers}[0];
-
-  use Mail::SpamAssassin::Constants qw(:ip);
-  my $ip64 = IP_ADDRESS;
-  my $ip4 = IPV4_ADDRESS;
-
-  # was the nameserver in IPv6 format?
-  if ($firstns =~ /^${ip64}$/o && $firstns !~ /^${ip4}$/o) {
-    my $addr = inet_aton($firstns);
-    if (!defined $addr) {
-      die "IO::Socket::INET6 module is required to use IPv6 nameservers such as '$firstns': $@\n";
-    }
-  }
-}
-
 1;
 
 =back
