@@ -524,12 +524,6 @@ sub finish {
   # Clean ourself up
   $self->finish_metadata();
 
-  # delete temporary files
-  if ($self->{'tmpfiles'}) {
-    unlink @{$self->{'tmpfiles'}};
-    delete $self->{'tmpfiles'};
-  }
-
   # These will only be in the root Message node
   delete $self->{'mime_boundary_state'};
   delete $self->{'mbox_sep'};
@@ -543,6 +537,17 @@ sub finish {
 
   # Go ahead and clean up all of the Message::Node parts
   while (my $part = shift @toclean) {
+    # bug 5557: windows requires tmp file be closed before it can be rm'd
+    if (ref $part->{'raw'} eq 'GLOB') {
+      close ($part->{'raw'});
+    }
+
+    # bug 5858: avoid memory leak with deep MIME structure
+    if (defined ($part->{metadata})) {
+      $part->{metadata}->finish();
+      delete $part->{metadata};
+    }
+
     delete $part->{'headers'};
     delete $part->{'raw_headers'};
     delete $part->{'header_order'};
@@ -560,6 +565,12 @@ sub finish {
       delete $part->{'body_parts'};
     }
   }
+
+  # delete temporary files
+  if ($self->{'tmpfiles'}) {
+    unlink @{$self->{'tmpfiles'}};
+    delete $self->{'tmpfiles'};
+  }
 }
 
 # also use a DESTROY method, just to ensure (as much as possible) that
@@ -634,7 +645,11 @@ sub parse_body {
       # If it's not multipart, go ahead and just deal with it.
       $self->_parse_normal($toparse);
 
-      if ($toparse->[0]->{'type'} =~ /^message\b/i && ($toparse->[3] > 0)) {
+      # bug 5041: exclude message/partial messages, however
+      if ($toparse->[0]->{'type'} =~ /^message\b/i &&
+          $toparse->[0]->{'type'} !~ /^message\/partial$/i &&
+            ($toparse->[3] > 0))
+      {
         # Just decode the part, but we don't care about the result here.
         $toparse->[0]->decode(0);
 
