@@ -331,7 +331,6 @@ sub decode {
         $self->{'decoded'} = Mail::SpamAssassin::Util::base64_decode($raw);
       }
 
-      # If it's a type text or message, split it into an array of lines
       if ( $self->{'type'} =~ m@^(?:text|message)\b/@i ) {
         $self->{'decoded'} =~ s/\015\012/\012/gs;
       }
@@ -688,7 +687,7 @@ sub get_all_headers {
 	$self->{'truncated_header'} = 1;
 	last HEADER;
       }
-      $lines[$position] = $self->{header_order}->[$position] . ": $contents";
+      $lines[$position] = $self->{header_order}->[$position].":".$contents;
     }
   }
 
