@@ -31,7 +31,19 @@ domain names from those, querying their NS records in DNS, resolving
 the hostnames used therein, and querying various DNS blocklists for
 those IP addresses.  This is quite effective.
 
-=head1 CONFIGURATION
+=head1 USER SETTINGS
+
+=over 4
+
+=item uridnsbl_skip_domain domain1 domain2 ...
+
+Specify a domain, or a number of domains, which should be skipped for the
+URIBL checks.  This is very useful to specify very common domains which are
+not going to be listed in URIBLs.
+
+=back
+
+=head1 RULE DEFINITIONS AND PRIVILEGED SETTINGS
 
 =over 4
 
@@ -88,24 +100,23 @@ Example:
   urirhssub   URIBL_RHSBL_4    rhsbl.example.org.   A    127.0.0.4
   urirhssub   URIBL_RHSBL_8    rhsbl.example.org.   A    8
 
-=item uridnsbl_timeout N		(default: 2)
+=back
 
-Specify the maximum number of seconds to wait for a result before
-giving up on the lookup.  Note that this is in addition to the normal
-DNS timeout applied for DNSBL lookups on IPs found in the Received headers.
+=head1 ADMINISTRATOR SETTINGS
+
+=over 4
 
 =item uridnsbl_max_domains N		(default: 20)
 
 The maximum number of domains to look up.
 
-=item uridnsbl_skip_domain domain1 domain2 ...
-
-Specify a domain, or a number of domains, which should be skipped for the
-URIBL checks.  This is very useful to specify very common domains which are
-not going to be listed in URIBLs.
-
 =back
 
+=head1 NOTES
+
+The C<uridnsbl_timeout> option has been obsoleted by the C<rbl_timeout>
+option.  See the C<Mail::SpamAssassin::Conf> POD for details on C<rbl_timeout>.
+
 =cut
 
 package Mail::SpamAssassin::Plugin::URIDNSBL;
@@ -159,33 +170,31 @@ sub parsed_metadata {
   if (!$scanner->is_dns_available()) {
     $self->{dns_not_available} = 1;
     return;
+  } else {
+    # due to re-testing dns may become available after being unavailable
+    # DOS: I don't think dns_not_available is even used anymore
+    $self->{dns_not_available} = 0;
   }
 
-  $self->{scanner} = $scanner;
-  my $scanstate = $scanner->{uribl_scanstate} = {
-    self => $self,
-    scanner => $scanner,
-    activerules => { },
-    hits => { }
-  };
+  $scanner->{'uridnsbl_activerules'} = { };
+  $scanner->{'uridnsbl_hits'} = { };
+  $scanner->{'uridnsbl_seen_domain'} = { };
 
   # only hit DNSBLs for active rules (defined and score != 0)
-  $scanstate->{active_rules_rhsbl} = { };
-  $scanstate->{active_rules_revipbl} = { };
+  $scanner->{'uridnsbl_active_rules_rhsbl'} = { };
+  $scanner->{'uridnsbl_active_rules_revipbl'} = { };
+
   foreach my $rulename (keys %{$scanner->{conf}->{uridnsbls}}) {
     next unless ($scanner->{conf}->is_rule_active('body_evals',$rulename));
 
-    my $rulecf = $scanstate->{scanner}->{conf}->{uridnsbls}->{$rulename};
+    my $rulecf = $scanner->{conf}->{uridnsbls}->{$rulename};
     if ($rulecf->{is_rhsbl}) {
-      $scanstate->{active_rules_rhsbl}->{$rulename} = 1;
+      $scanner->{uridnsbl_active_rules_rhsbl}->{$rulename} = 1;
     } else {
-      $scanstate->{active_rules_revipbl}->{$rulename} = 1;
+      $scanner->{uridnsbl_active_rules_revipbl}->{$rulename} = 1;
     }
   }
 
-  $self->setup ($scanstate);
-
-
   # get all domains in message
 
   # don't keep dereferencing this
@@ -247,7 +256,8 @@ sub parsed_metadata {
   # at this point, @uri_ordered is an ordered array of uri hashes
 
   my %domlist = ();
-  while (keys %domlist < $scanner->{main}->{conf}->{uridnsbl_max_domains} && @uri_ordered) {
+  my $umd = $scanner->{main}->{conf}->{uridnsbl_max_domains};
+  while (keys %domlist < $umd && @uri_ordered) {
     my $array = shift @uri_ordered;
     next unless $array;
 
@@ -256,7 +266,7 @@ sub parsed_metadata {
     next unless @domains;
 
     # the new domains are all useful, just add them in
-    if (keys(%domlist) + @domains <= $scanner->{main}->{conf}->{uridnsbl_max_domains}) {
+    if (keys(%domlist) + @domains <= $umd) {
       foreach (@domains) {
         $domlist{$_} = 1;
       }
@@ -264,7 +274,7 @@ sub parsed_metadata {
     else {
       # trim down to a limited number - pick randomly
       my $i;
-      while (@domains && keys %domlist < $scanner->{main}->{conf}->{uridnsbl_max_domains}) {
+      while (@domains && keys %domlist < $umd) {
         my $r = int rand (scalar @domains);
         $domlist{splice (@domains, $r, 1)} = 1;
       }
@@ -274,7 +284,7 @@ sub parsed_metadata {
   # and query
   dbg("uridnsbl: domains to query: ".join(' ',keys %domlist));
   foreach my $dom (keys %domlist) {
-    $self->query_domain ($scanstate, $dom);
+    $self->query_domain ($scanner, $dom);
   }
 
   return 1;
@@ -284,20 +294,16 @@ sub set_config {
   my($self, $conf) = @_;
   my @cmds = ();
 
-  push(@cmds, {
-    setting => 'uridnsbl_timeout',
-    default => 3,
-    type => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
-  });
-
   push(@cmds, {
     setting => 'uridnsbl_max_domains',
+    is_admin => 1,
     default => 20,
     type => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
   });
 
   push (@cmds, {
     setting => 'uridnsbl',
+    is_priv => 1,
     code => sub {
       my ($self, $key, $value, $line) = @_;
       if ($value =~ /^(\S+)\s+(\S+)\s+(\S+)$/) {
@@ -320,6 +326,7 @@ sub set_config {
 
   push (@cmds, {
     setting => 'urirhsbl',
+    is_priv => 1,
     code => sub {
       my ($self, $key, $value, $line) = @_;
       if ($value =~ /^(\S+)\s+(\S+)\s+(\S+)$/) {
@@ -342,6 +349,7 @@ sub set_config {
 
   push (@cmds, {
     setting => 'urirhssub',
+    is_priv => 1,
     code => sub {
       my ($self, $key, $value, $line) = @_;
       if ($value =~ /^(\S+)\s+(\S+)\s+(\S+)\s+(\d{1,10}|\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})$/) {
@@ -354,9 +362,7 @@ sub set_config {
           is_rhsbl => 1, is_subrule => 1
         };
         $self->{uridnsbl_subs}->{$zone} ||= { };
-        $self->{uridnsbl_subs}->{$zone}->{$subrule} = {
-          rulename => $rulename
-        };
+        push (@{$self->{uridnsbl_subs}->{$zone}->{$subrule}->{rulenames}}, $rulename);
       }
       elsif ($value =~ /^$/) {
         return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
@@ -381,77 +387,31 @@ sub set_config {
     }
   });
 
+  # obsolete
+  push(@cmds, {
+    setting => 'uridnsbl_timeout',
+    code => sub {
+      # not a lint_warn(), since it's pretty harmless and we don't want
+      # to break stuff like sa-update
+      warn("config: 'uridnsbl_timeout' is obsolete, use 'rbl_timeout' instead");
+      return 0;
+    }
+  });
+
   $conf->{parser}->register_commands(\@cmds);
 }
 
-sub check_tick {
-  my ($self, $opts) = @_;
-
-  return if ($self->{dns_not_available});
-  $self->complete_lookups($opts->{permsgstatus}->{uribl_scanstate}, 0.3);
-  return 1;
-}
-
-sub check_post_dnsbl {
-  my ($self, $opts) = @_;
-
-  return if ($self->{dns_not_available});
-
-  my $scan = $opts->{permsgstatus};
-  my $scanstate = $scan->{uribl_scanstate};
-
-  # try to complete a few more
-  if (!$self->complete_lookups($scanstate, 0.1)) {
-    my $secs_to_wait = $scan->{conf}->{uridnsbl_timeout};
-    if ($secs_to_wait < 0) { $secs_to_wait = 0; }
-    my $now = time;
-    my $deadline = $now + $secs_to_wait;
-    dbg("uridnsbl: waiting $secs_to_wait seconds for URIDNSBL lookups to complete");
-
-    while ($now < $deadline) {
-      last if ($self->complete_lookups($scanstate, $deadline - $now));
-      $now = time;
-    }
-    dbg("uridnsbl: done waiting for URIDNSBL lookups to complete");
-  }
-
-  foreach my $rulename (keys %{$scanstate->{active_rules_revipbl}},
-                        keys %{$scanstate->{active_rules_rhsbl}})
-  {
-    $scan->clear_test_state();
-
-    if ($scanstate->{hits}->{$rulename}) {
-      my $uris = join (' ', keys %{$scanstate->{hits}->{$rulename}});
-      $scan->test_log ("URIs: $uris");
-      $scan->got_hit ($rulename, "");
-    }
-  }
-
-  $self->abort_remaining_lookups ($scanstate);
-}
-
-# ---------------------------------------------------------------------------
-
-sub setup {
-  my ($self, $scanstate) = @_;
-
-  $scanstate->{pending_lookups} = { };
-  $scanstate->{seen_domain} = { };
-  $scanstate->{last_count} = 0;
-  $scanstate->{times_count_was_same} = 0;
-}
-
 # ---------------------------------------------------------------------------
 
 sub query_domain {
-  my ($self, $scanstate, $dom) = @_;
+  my ($self, $scanner, $dom) = @_;
 
   #warn "uridnsbl: domain $dom\n";
   #return;
 
   $dom = lc $dom;
-  return if $scanstate->{seen_domain}->{$dom};
-  $scanstate->{seen_domain}->{$dom} = 1;
+  return if $scanner->{uridnsbl_seen_domain}->{$dom};
+  $scanner->{uridnsbl_seen_domain}->{$dom} = 1;
   $self->log_dns_result("querying domain $dom");
 
   my $obj = {
@@ -460,12 +420,12 @@ sub query_domain {
   };
 
   my $single_dnsbl = 0;
-  if ($dom =~ /^\d+\.\d+\.\d+\.\d+$/) { 
+  if ($dom =~ /^\d+\.\d+\.\d+\.\d+$/) {
     my $IPV4_ADDRESS = IPV4_ADDRESS;
     my $IP_PRIVATE = IP_PRIVATE;
     # only look up the IP if it is public and valid
     if ($dom =~ /^$IPV4_ADDRESS$/ && $dom !~ /^$IP_PRIVATE$/) {
-      $self->lookup_dnsbl_for_ip($scanstate, $obj, $dom);
+      $self->lookup_dnsbl_for_ip($scanner, $obj, $dom);
       # and check the IP in RHSBLs too
       if ($dom =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/) {
 	$dom = "$4.$3.$2.$1";
@@ -479,39 +439,51 @@ sub query_domain {
 
   if ($single_dnsbl) {
     # look up the domain in the RHSBL subset
-    my $cf = $scanstate->{active_rules_rhsbl};
+    my $cf = $scanner->{uridnsbl_active_rules_rhsbl};
     foreach my $rulename (keys %{$cf}) {
-      my $rulecf = $scanstate->{scanner}->{conf}->{uridnsbls}->{$rulename};
-      $self->lookup_single_dnsbl($scanstate, $obj, $rulename,
+      my $rulecf = $scanner->{conf}->{uridnsbls}->{$rulename};
+      $self->lookup_single_dnsbl($scanner, $obj, $rulename,
 				 $dom, $rulecf->{zone}, $rulecf->{type});
+
+      # see comment below
+      $scanner->register_async_rule_start($rulename);
     }
 
     # perform NS, A lookups to look up the domain in the non-RHSBL subset
     if ($dom !~ /^\d+\.\d+\.\d+\.\d+$/) {
-      $self->lookup_domain_ns($scanstate, $obj, $dom);
+      $self->lookup_domain_ns($scanner, $obj, $dom);
     }
   }
+
+  # note that these rules are now underway.   important: unless the
+  # rule hits, in the current design, these will not be considered
+  # "finished" until harvest_dnsbl_queries() completes
+  my $cf = $scanner->{uridnsbl_active_rules_revipbl};
+  foreach my $rulename (keys %{$cf}) {
+    $scanner->register_async_rule_start($rulename);
+  }
 }
 
 # ---------------------------------------------------------------------------
 
 sub lookup_domain_ns {
-  my ($self, $scanstate, $obj, $dom) = @_;
+  my ($self, $scanner, $obj, $dom) = @_;
 
   my $key = "NS:".$dom;
-  return if $scanstate->{pending_lookups}->{$key};
+  return if $scanner->{async}->get_lookup($key);
 
   # dig $dom ns
-  my $ent = $self->start_lookup ($scanstate, 'NS', $self->res_bgsend($dom, 'NS'));
+  my $ent = $self->start_lookup($scanner, $dom, 'NS',
+                                $self->res_bgsend($scanner, $dom, 'NS', $key),
+                                $key);
   $ent->{obj} = $obj;
-  $scanstate->{pending_lookups}->{$key} = $ent;
 }
 
 sub complete_ns_lookup {
-  my ($self, $scanstate, $ent, $dom) = @_;
+  my ($self, $scanner, $ent, $dom) = @_;
 
   my $packet = $ent->{response_packet};
-  my @answer = $packet->answer;
+  my @answer = !defined $packet ? () : $packet->answer;
 
   my $IPV4_ADDRESS = IPV4_ADDRESS;
   my $IP_PRIVATE = IP_PRIVATE;
@@ -528,11 +500,11 @@ sub complete_ns_lookup {
 	$nsmatch =~ s/\.$//;
 	# only look up the IP if it is public and valid
 	if ($nsmatch =~ /^$IPV4_ADDRESS$/ && $nsmatch !~ /^$IP_PRIVATE$/) {
-	  $self->lookup_dnsbl_for_ip($scanstate, $ent->{obj}, $nsmatch);
+	  $self->lookup_dnsbl_for_ip($scanner, $ent->{obj}, $nsmatch);
 	}
       }
       else {
-	$self->lookup_a_record($scanstate, $ent->{obj}, $nsmatch);
+	$self->lookup_a_record($scanner, $ent->{obj}, $nsmatch);
       }
     }
   }
@@ -541,26 +513,29 @@ sub complete_ns_lookup {
 # ---------------------------------------------------------------------------
 
 sub lookup_a_record {
-  my ($self, $scanstate, $obj, $hname) = @_;
+  my ($self, $scanner, $obj, $hname) = @_;
 
   my $key = "A:".$hname;
-  return if $scanstate->{pending_lookups}->{$key};
+  return if $scanner->{async}->get_lookup($key);
 
   # dig $hname a
-  my $ent = $self->start_lookup ($scanstate, 'A', $self->res_bgsend($hname, 'A'));
+  my $ent = $self->start_lookup($scanner, $hname, 'A',
+                                $self->res_bgsend($scanner, $hname, 'A', $key),
+                                $key);
   $ent->{obj} = $obj;
-  $scanstate->{pending_lookups}->{$key} = $ent;
 }
 
 sub complete_a_lookup {
-  my ($self, $scanstate, $ent, $hname) = @_;
+  my ($self, $scanner, $ent, $hname) = @_;
 
-  foreach my $rr ($ent->{response_packet}->answer) {
+  my $packet = $ent->{response_packet};
+  my @answer = !defined $packet ? () : $packet->answer;
+  foreach my $rr (@answer) {
     my $str = $rr->string;
     $self->log_dns_result ("A for NS $hname: $str");
 
     if ($str =~ /IN\s+A\s+(\S+)/) {
-      $self->lookup_dnsbl_for_ip($scanstate, $ent->{obj}, $1);
+      $self->lookup_dnsbl_for_ip($scanner, $ent->{obj}, $1);
     }
   }
 }
@@ -568,46 +543,45 @@ sub complete_a_lookup {
 # ---------------------------------------------------------------------------
 
 sub lookup_dnsbl_for_ip {
-  my ($self, $scanstate, $obj, $ip) = @_;
+  my ($self, $scanner, $obj, $ip) = @_;
 
   $ip =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/;
   my $revip = "$4.$3.$2.$1";
 
-  my $cf = $scanstate->{active_rules_revipbl};
+  my $cf = $scanner->{uridnsbl_active_rules_revipbl};
   foreach my $rulename (keys %{$cf}) {
-    my $rulecf = $scanstate->{scanner}->{conf}->{uridnsbls}->{$rulename};
-    $self->lookup_single_dnsbl($scanstate, $obj, $rulename,
+    my $rulecf = $scanner->{conf}->{uridnsbls}->{$rulename};
+    $self->lookup_single_dnsbl($scanner, $obj, $rulename,
 			       $revip, $rulecf->{zone}, $rulecf->{type});
   }
 }
 
 sub lookup_single_dnsbl {
-  my ($self, $scanstate, $obj, $rulename, $lookupstr, $dnsbl, $qtype) = @_;
+  my ($self, $scanner, $obj, $rulename, $lookupstr, $dnsbl, $qtype) = @_;
 
   my $key = "DNSBL:".$dnsbl.":".$lookupstr;
-  return if $scanstate->{pending_lookups}->{$key};
+  return if $scanner->{async}->get_lookup($key);
   my $item = $lookupstr.".".$dnsbl;
 
   # dig $ip txt
-  my $ent = $self->start_lookup ($scanstate, 'DNSBL',
-        $self->res_bgsend($item, $qtype));
+  my $ent = $self->start_lookup($scanner, $item, 'DNSBL',
+                              $self->res_bgsend($scanner, $item, $qtype, $key),
+                              $key);
   $ent->{obj} = $obj;
   $ent->{rulename} = $rulename;
   $ent->{zone} = $dnsbl;
-  $scanstate->{pending_lookups}->{$key} = $ent;
 }
 
 sub complete_dnsbl_lookup {
-  my ($self, $scanstate, $ent, $dnsblip) = @_;
+  my ($self, $scanner, $ent, $dnsblip) = @_;
 
-  my $scan = $scanstate->{scanner};
-  my $conf = $scan->{conf};
+  my $conf = $scanner->{conf};
   my @subtests = ();
   my $rulename = $ent->{rulename};
   my $rulecf = $conf->{uridnsbls}->{$rulename};
 
   my $packet = $ent->{response_packet};
-  my @answer = $packet->answer;
+  my @answer = !defined $packet ? () : $packet->answer;
 
   my $uridnsbl_subs = $conf->{uridnsbl_subs}->{$ent->{zone}};
   foreach my $rr (@answer)
@@ -625,22 +599,24 @@ sub complete_dnsbl_lookup {
             $packet->header->id." rr=".$rr->string);
 	next;
       }
-      $self->got_dnsbl_hit($scanstate, $ent, $rdatastr, $dom, $rulename);
+      $self->got_dnsbl_hit($scanner, $ent, $rdatastr, $dom, $rulename);
     }
     else {
       foreach my $subtest (keys (%{$uridnsbl_subs}))
       {
-        my $subrulename = $uridnsbl_subs->{$subtest}->{rulename};
-
         if ($subtest eq $rdatastr) {
-          $self->got_dnsbl_hit($scanstate, $ent, $rdatastr, $dom, $subrulename);
+          foreach my $subrulename (@{$uridnsbl_subs->{$subtest}->{rulenames}}) {
+            $self->got_dnsbl_hit($scanner, $ent, $rdatastr, $dom, $subrulename);
+          }
         }
         # bitmask
         elsif ($subtest =~ /^\d+$/) {
 	  if ($rdatastr =~ m/^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/ &&
 	      Mail::SpamAssassin::Util::my_inet_aton($rdatastr) & $subtest)
           {
-            $self->got_dnsbl_hit($scanstate, $ent, $rdatastr, $dom, $subrulename);
+            foreach my $subrulename (@{$uridnsbl_subs->{$subtest}->{rulenames}}) {
+              $self->got_dnsbl_hit($scanner, $ent, $rdatastr, $dom, $subrulename);
+            }
           }
         }
       }
@@ -649,152 +625,76 @@ sub complete_dnsbl_lookup {
 }
 
 sub got_dnsbl_hit {
-  my ($self, $scanstate, $ent, $str, $dom, $rulename) = @_;
+  my ($self, $scanner, $ent, $str, $dom, $rulename) = @_;
 
   $str =~ s/\s+/  /gs;	# long whitespace => short
   dbg("uridnsbl: domain \"$dom\" listed ($rulename): $str");
 
-  if (!defined $scanstate->{hits}->{$rulename}) {
-    $scanstate->{hits}->{$rulename} = { };
+  if (!defined $scanner->{uridnsbl_hits}->{$rulename}) {
+    $scanner->{uridnsbl_hits}->{$rulename} = { };
   };
-  $scanstate->{hits}->{$rulename}->{$dom} = 1;
+  $scanner->{uridnsbl_hits}->{$rulename}->{$dom} = 1;
+
+  if ($scanner->{uridnsbl_active_rules_revipbl}->{$rulename}
+    || $scanner->{uridnsbl_active_rules_rhsbl}->{$rulename})
+  {
+    # TODO: this needs to handle multiple domain hits per rule
+    $scanner->clear_test_state();
+    my $uris = join (' ', keys %{$scanner->{uridnsbl_hits}->{$rulename}});
+    $scanner->test_log ("URIs: $uris");
+    $scanner->got_hit ($rulename, "");
+
+    # note that this rule has completed (since it got at least 1 hit)
+    $scanner->register_async_rule_finish($rulename);
+  }
 }
 
 # ---------------------------------------------------------------------------
 
 sub start_lookup {
-  my ($self, $scanstate, $type, $id) = @_;
+  my ($self, $scanner, $zone, $type, $id, $key) = @_;
+
   my $ent = {
-    type => $type,
-    id => $id
-  };
-  $scanstate->{queries_started}++;
-  $ent;
-}
-
-# ---------------------------------------------------------------------------
-
-# perform a poll of our lookups, to see if any are completed; if they
-# are, the next lookup in the sequence will be kicked off.
-
-sub complete_lookups {
-  my ($self, $scanstate, $timeout) = @_;
-  my %typecount = ();
-  my $stillwaiting = 0;
-
-  my $pending = $scanstate->{pending_lookups};
-  if (scalar keys %{$pending} <= 0) {
-    return 1;		# nothing left to do
-  }
-
-  $scanstate->{queries_started} = 0;
-  $scanstate->{queries_completed} = 0;
-
-  my $nfound = $self->{main}->{resolver}->poll_responses($timeout);
-  $nfound ||= 'no';
-  dbg ("uridnsbl: select found $nfound socks ready");
-
-  foreach my $key (keys %{$pending}) {
-    my $ent = $pending->{$key};
-    my $type = $ent->{type};
-
-    if (!exists ($self->{finished}->{$ent->{id}})) {
-      $typecount{$type}++;
-      #$stillwaiting = 1;
-      next;
-    }
-
-    $ent->{response_packet} = delete $self->{finished}->{$ent->{id}};
-    $key =~ /:(\S+)$/; my $val = $1;
-
-    if (LOG_COMPLETION_TIMES) {
-      my $secs = (time - $ent->{start});
-      my $totalsecs = (time - $ent->{obj}->{querystart});
-      printf "# time: %s %3.3f %3.3f %s\n",
-		$type, $secs, $totalsecs, $ent->{obj}->{dom};
-    }
-
-    if ($type eq 'NS') {
-      $self->complete_ns_lookup ($scanstate, $ent, $val);
-    }
-    elsif ($type eq 'A') {
-      $self->complete_a_lookup ($scanstate, $ent, $val);
-    }
-    elsif ($type eq 'DNSBL') {
-      $self->complete_dnsbl_lookup ($scanstate, $ent, $val);
-      my $totalsecs = (time - $ent->{obj}->{querystart});
-      dbg("uridnsbl: query for ".$ent->{obj}->{dom}." took ".
-		$totalsecs." seconds to look up ($val)");
-    }
-
-    $scanstate->{queries_completed}++;
-    delete $scanstate->{pending_lookups}->{$key};
-  }
-
-  dbg("uridnsbl: queries completed: ".$scanstate->{queries_completed}.
-		" started: ".$scanstate->{queries_started});
-
-  if (1) {
-    dbg("uridnsbl: queries active: ".
-	join (' ', map { "$_=$typecount{$_}" } sort keys %typecount)." at ".
-	localtime(time));
-  }
-
-  # ensure we don't get stuck if a request gets lost in the ether.
-  if (!$stillwaiting) {
-    my $numkeys = scalar keys %{$scanstate->{pending_lookups}};
-    if ($numkeys == 0) {
-      $stillwaiting = 0;
-
-    } else {
-      $stillwaiting = 1;
-
-      # avoid looping forever if we haven't got all results. 
-      if ($scanstate->{last_count} == $numkeys) {
-	$scanstate->{times_count_was_same}++;
-	if ($scanstate->{times_count_was_same} > 20) {
-	  dbg("uridnsbl: escaping: must have lost requests");
-	  $self->abort_remaining_lookups ($scanstate);
-	  $stillwaiting = 0;
-	}
-      } else {
-	$scanstate->{last_count} = $numkeys;
-	$scanstate->{times_count_was_same} = 0;
+    key => $key,
+    zone => $zone,  # serves to fetch other per-zone settings
+    type => "URI-".$type,
+    id => $id,
+    completed_callback => sub {
+      my $ent = shift;
+      if (defined $ent->{response_packet}) {  # not aborted or empty
+        $self->completed_lookup_callback ($scanner, $ent);
       }
     }
-  }
-
-  return (!$stillwaiting);
+  };
+  $scanner->{async}->start_lookup($ent);
+  return $ent;
 }
 
-# ---------------------------------------------------------------------------
+sub completed_lookup_callback {
+  my ($self, $scanner, $ent) = @_;
+  my $type = $ent->{type};
+  my $key = $ent->{key};
+  $key =~ /:(\S+?)$/; my $val = $1;
 
-sub abort_remaining_lookups  {
-  my ($self, $scanstate) = @_;
-
-  my $pending = $scanstate->{pending_lookups};
-  my $foundone = 0;
-  foreach my $key (keys %{$pending})
-  {
-    if (!$foundone) {
-      dbg("uridnsbl: aborting remaining lookups");
-      $foundone = 1;
-    }
-
-    delete $pending->{$key};
+  if ($type eq 'URI-NS') {
+    $self->complete_ns_lookup ($scanner, $ent, $val);
+  }
+  elsif ($type eq 'URI-A') {
+    $self->complete_a_lookup ($scanner, $ent, $val);
+  }
+  elsif ($type eq 'URI-DNSBL') {
+    $self->complete_dnsbl_lookup ($scanner, $ent, $val);
   }
-  $self->{main}->{resolver}->bgabort();
 }
 
 # ---------------------------------------------------------------------------
 
 sub res_bgsend {
-  my ($self, $host, $type) = @_;
+  my ($self, $scanner, $host, $type, $key) = @_;
 
   return $self->{main}->{resolver}->bgsend($host, $type, undef, sub {
-        my $pkt = shift;
-        my $id = shift;
-        $self->{finished}->{$id} = $pkt;
+        my ($pkt, $id, $timestamp) = @_;
+        $scanner->{async}->set_response_packet($id, $pkt, $key, $timestamp);
       });
 }
 
