@@ -36,12 +36,22 @@ my @MODULES = (
 },
 {
   'module' => 'HTML::Parser',
-  'version' => '3.24',
-  'desc' => 'Version 3.31 or later is recommended.
-
-  HTML is used for an ever-increasing amount of email so this dependency
+  'version' => '3.43',
+  'desc' => 'HTML is used for an ever-increasing amount of email so this dependency
   is unavoidable.  Run "perldoc -q html" for additional information.',
 },
+{
+  module => 'Net::DNS',
+  version => ($^O =~ /^(mswin|dos|os2)/oi ? '0.46' : '0.34'),
+  desc => 'Used for all DNS-based tests (SBL, XBL, SpamCop, DSBL, etc.),
+  perform MX checks, and is also used when manually reporting spam to
+  SpamCop.
+
+  You need to make sure the Net::DNS version is sufficiently up-to-date:
+
+  - version 0.34 or higher on Unix systems
+  - version 0.46 or higher on Windows systems',
+},
 );
 
 my @OPTIONAL_MODULES = (
@@ -58,31 +68,25 @@ my @OPTIONAL_MODULES = (
   auto-whitelist.  *Much* more efficient than the other standard Perl
   database packages.  Strongly recommended.',
 },
-{
-  module => 'Net::DNS',
-  version => ($^O =~ /^(mswin|dos|os2)/oi ? '0.46' : '0.34'),
-  desc => 'Used for all DNS-based tests (SBL, XBL, SpamCop, DSBL, etc.),
-  perform MX checks, and is also used when manually reporting spam to
-  SpamCop.  Recommended.
-
-  If this is installed and you are using network tests of any variety
-  (which is the default), then you need to make sure the Net::DNS
-  version is sufficiently up-to-date:
-
-  - version 0.34 or higher on Unix systems
-  - version 0.46 or higher on Windows systems',
-},
 {
   module => 'Net::SMTP',
   alt_name => 'libnet',
   version => '0.00',
   desc => 'Used when manually reporting spam to SpamCop with "spamassassin -r".',
 },
+{
+  module => 'Mail::SPF',
+  version => '0.00',
+  desc => 'Used to check DNS Sender Policy Framework (SPF) records to fight email
+  address forgery and make it easier to identify spams.  (This is preferred
+  over Mail::SPF::Query.)',
+},
 {
   module => 'Mail::SPF::Query',
   version => '0.00',
   desc => 'Used to check DNS Sender Policy Framework (SPF) records to fight email
-  address forgery and make it easier to identify spams.',
+  address forgery and make it easier to identify spams.  (Mail::SPF is
+  preferred instead of this module.)',
 },
 {
   module => 'IP::Country::Fast',
@@ -125,11 +129,35 @@ my @OPTIONAL_MODULES = (
   ENABLE_SSL="yes" argument to Makefile.PL to build and run an SSL
   compatibile spamc.)',
 },
+{
+  module => 'Compress::Zlib',
+  version => '0.00',
+  desc => 'If you wish to use the optional zlib compression for communication
+  between spamc and spamd (the -z option to spamc), you need to install
+  this module.',
+},
 {
   module => 'Time::HiRes',
   version => '0.00',
-  desc => 'If this module is installed, the processing times are logged/reported
-  more precisely in spamd.',
+  desc => 'If this module is installed, asynchronous DNS lookup timeouts operate
+  with subsecond precision and the processing times are logged/reported
+  more accurately. Other modules and plugins may benefit too.',
+},
+{
+  module => 'Mail::DomainKeys',
+  version => '0.00',
+  desc => 'If this module is installed, and you enable the DomainKeys plugin,
+  SpamAssassin will perform Domain Key lookups when Domain Key
+  information is present in the message headers.  (Note that new versions
+  of Mail::DKIM render this module superfluous.)'
+},
+{
+  module => 'Mail::DKIM',
+  version => '0.00',
+  desc => 'If this module is installed, and you enable the DKIM plugin,
+  SpamAssassin will perform DKIM lookups when a DKIM-Signature
+  header is present in the message headers.  (New versions of this module
+  support both Domain Keys and DKIM, rendering Mail::DomainKeys obsolete.)'
 },
 {
   module => 'DBI',
@@ -169,6 +197,13 @@ my @OPTIONAL_MODULES = (
   desc => 'The "sa-update" script requires this module to access compressed
   update archive files.',
 },
+{
+  module => 'Encode::Detect',
+  version => '0.00',
+  desc => 'If you plan to use the normalize_charset config setting to detect
+  charsets and convert them into Unicode, you will need to install
+  this module.',
+},
 );
 
 ###########################################################################
@@ -191,7 +226,7 @@ sub debug_diagnostics {
     sub Net::Ident::_export_hooks;
   ';
 
-  foreach my $moddef (sort (@MODULES, @OPTIONAL_MODULES)) {
+  foreach my $moddef (@MODULES, @OPTIONAL_MODULES) {
     my $module = $moddef->{module};
     my $modver;
     if (eval ' require '.$module.'; $modver = $'.$module.'::VERSION; 1;')
@@ -256,7 +291,7 @@ sub try_module {
   print "\n", ("*" x 75), "\n";
   if ($required) {
     $EXIT_STATUS++;
-    print "\aERROR: the required $pretty_name ${pretty_version}module $errtype";
+    warn "\aERROR: the required $pretty_name ${pretty_version}module $errtype";
     if ($not_installed) {
       $$summref .= "REQUIRED module missing: $pretty_name\n";
     } else {
@@ -275,3 +310,5 @@ sub try_module {
 
   print "\n\n".$desc."\n\n";
 }
+
+1;
