@@ -26,7 +26,9 @@ CREATE TABLE bayes_token (
   spam_count int(11) NOT NULL default '0',
   ham_count int(11) NOT NULL default '0',
   atime int(11) NOT NULL default '0',
-  PRIMARY KEY  (id, token)
+  PRIMARY KEY  (id, token),
+  INDEX bayes_token_idx1 (token),
+  INDEX bayes_token_idx2 (id, atime)
 ) TYPE=MyISAM;
 
 CREATE TABLE bayes_vars (
